from pathlib import Path
from typing import Union

import orjson
from typeguard import typechecked


@typechecked
def read_json(filename: Union[str, Path]):
    with open(filename, "rb") as f:
        return orjson.loads(f.read())



from functools import cached_property
from urllib.request import urlretrieve

import pandas as pd

from .._io import read_json
from .._table import EntityTable


class Disease(EntityTable):
    """Disease.

    Edits of terms are coordinated and reviewed on:
    https://github.com/monarch-initiative/mondo
    """

    def __init__(self, reload: bool = False) -> None:
        filename, _ = urlretrieve(
            "https://bionty-assets.s3.amazonaws.com/mondo-base.json"
        )
        self._onto_dict = read_json(filename)

    @cached_property
    def df(self) -> pd.DataFrame:
        """DataFrame."""
        return pd.DataFrame(pd.Series(self._onto_dict))



import numpy as np
import pandas as pd


def check_if_index_compliant(index: pd.Index, ref_index: pd.Index) -> np.ndarray:
    """The index already theoretically conforms with the Bionty ID for the entity.

    Meaning, the name of the index column is, for instance `hgnc_symbol`.

    Hence, this function only checks whether the terms in the index are all
    contained in the Bionty ID vocabulary.

    It returns a `pd.Index` of terms that aren't contained in the reference.
    """
    # boolean vector indicating standardized terms
    matches = index.isin(ref_index)
    return matches


def get_compliant_index_from_column(
    df: pd.DataFrame,
    ref_df: pd.DataFrame,
    column: str,
):
    """Get a reference-ID-compliant index based on a column with an alternative identifier.

    Example:
    >>> compliant_index = map_index_from_column(df, "ensemble_id")
    >>> df.index = compliant_index

    This will fail for some entries if the column is not uniquely mappable to
    the Bionty ID.

    There are two options for the user:
    1. Maintain previous index identifiers for the terms that aren't
    mappable and flag them (`keep_data == True`)
    2. Reduce the dimensions to the ones that are mappable (`keep_data == False`)

    Args:
        df: DataFrame with an index different from reference ID, but with a
            column that has mappable information, like an ensemble_id.
        ref_df: Reference EntityTable in Bionty.
        column: Column to be mapped to reference ID.
        keep_data: Keep terms that are not mappable to the reference ID.
    """
    df = df.copy()  # not touch the input df

    # prepare ref_df for lookup
    ref_index = ref_df.index
    ref_df = ref_df.reset_index()

    if column not in ref_df.columns:
        raise AssertionError(f"{column} name must match one of {ref_df.columns}!")
    else:
        query_values = df[column].values

    # lookup_df is indexed with the query column field in the ref_df
    # and with a "bionty_id" column containing the Bionty default identifier
    # e.g. default is `hgnc_symbol` for human
    lookup_df = pd.DataFrame(index=ref_df[column], data={"bionty_id": ref_index})

    matches = check_if_index_compliant(pd.Index(query_values), lookup_df.index)

    if all(~matches):  # nothing is mappable
        raise AssertionError(f"{column} name must contain at least one mappable term!")

    # mapped_dict is the {query_id: bionty_id}, where query_id = ref_df[column]
    lookup_df_mappable = lookup_df.loc[query_values[matches]]
    mapper = lookup_df_mappable["bionty_id"]
    mapper = mapper.drop_duplicates()
    new_index = df[column].map(mapper)  # what's unmappable results in nan
    new_index = new_index.fillna(df[column]).values  # fill nans with orig values

    return pd.Index(new_index), matches


def explode_aggregated_column_to_expand(
    df: pd.DataFrame, aggregated_col=str, target_col=None, sep="|"
):
    """Explode values from an aggregated DataFrame column to expand a target column.

    Args:
        df: a DataFrame containing the aggregated_col and target_col
        aggregated_col: the name of the aggregated column
        target_col: the name of the target column
            If None, use the index as the target column

    Returns:
        a DataFrame index by the split values from the aggregated column;
        the target column is aggregated so that the new index is unique.
    """
    if target_col is None:
        # take the index as the target column
        if df.index.name is None:
            target_col = df.index.name = "index"
        else:
            target_col = df.index.name
    df = df.reset_index()[[aggregated_col, target_col]].copy()

    # explode the values from the aggregated cells into new rows
    df[aggregated_col] = df[aggregated_col].str.split(sep)
    exploded_df = df.explode(aggregated_col)

    # if any values in the aggregated column is already in the target col
    # sets those values of the aggregated column to None
    exploded_df.loc[
        exploded_df[aggregated_col].isin(exploded_df[target_col]), aggregated_col
    ] = None

    # set the values aggregated_col equal the target_col if None before concat
    exploded_df[aggregated_col] = exploded_df[aggregated_col].fillna(
        exploded_df[target_col]
    )

    # append the additional values in the target column to the df
    add_values = exploded_df[
        ~exploded_df[target_col].isin(exploded_df[aggregated_col])
    ][target_col].unique()
    add_df = pd.DataFrame(data={target_col: add_values, aggregated_col: add_values})

    # aggregate the target column so that the new index (aggregated column) is unique
    df_concat = pd.concat([exploded_df, add_df])
    df_concat = df_concat.groupby(aggregated_col).agg(sep.join)

    return df_concat



from functools import cached_property

import pandas as pd

from .._settings import check_datasetdir_exists, settings


class Protein:
    """Protein.

    Args:
        species: `common_name` of `Species` entity EntityTable.
    """

    def __init__(self, species="human") -> None:
        self._species = species
        self._filepath = settings.datasetdir / f"{self.species}-uniprot.feather"

    @property
    def entity(self):
        """Name of the entity."""
        return "protein"

    @property
    def species(self):
        """The `common_name` of `Species` entity EntityTable."""
        return self._species

    @property
    def filepath(self):
        """The local filepath to the DataFrame."""
        return self._filepath

    @cached_property
    def df(self):
        """DataFrame."""
        if self.species not in {"human", "mouse"}:
            raise NotImplementedError
        else:
            if not self.filepath.exists():
                self._download_df()
            return pd.read_feather(self.filepath).set_index("UniProtKB-AC")

    @check_datasetdir_exists
    def _download_df(self):
        from urllib.request import urlretrieve

        urlretrieve(
            f"https://bionty-assets.s3.amazonaws.com/{self.species}-uniprot.feather",
            self.filepath,
        )



import pandas as pd

from bionty import Gene

data = {
    "hgnc_symbol": ["A1CF", "A1BG", "corrupted1", "corrupted2"],
    "hgnc_id": ["HGNC:24086", "HGNC:5", "corrupted1", "corrupted2"],
    "ensembl.gene_id": [
        "ENSG00000148584",
        "ENSG00000121410",
        "corrupted1",
        "corrupted2",
    ],
}
df = pd.DataFrame(data).set_index("hgnc_id")


def test_curate():
    df_curated = Gene().curate(df.set_index("hgnc_symbol"))
    assert df_curated.index.to_list() == ["A1CF", "A1BG", "corrupted1", "corrupted2"]
    assert df_curated.__curated__.to_list() == [True, True, False, False]



from collections import namedtuple
from functools import cached_property

import pandas as pd

from .._normalize import GENE_COLUMNS, NormalizeColumns
from .._settings import check_datasetdir_exists, settings
from .._table import EntityTable

STD_ID_DICT = {"human": "hgnc_symbol", "mouse": "mgi_symbol"}
FILENAMES = {"human": "hgnc_complete_set.feather", "mouse": "mgi_complete_set.feather"}
ALIAS_DICT = {"hgnc_symbol": "alias_symbol", "mgi_symbol": "Synonyms"}


class Gene(EntityTable):
    """Gene.

    The default indexes chosen are
    - human: hgnc_symbol
    - mouse: mgi_symbol

    We think these identifiers are the best unambiguous ways to reference genes.

    Args:
        species: `common_name` of `Species` entity EntityTable.
        id: If `None`, chooses an id field in a species dependent way.

    Notes:
        Biotypes: https://useast.ensembl.org/info/genome/genebuild/biotypes.html
        Gene Naming: https://useast.ensembl.org/info/genome/genebuild/gene_names.html

    """

    def __init__(
        self,
        species="human",
        id=None,
    ):
        self._species = species
        self._filepath = settings.datasetdir / FILENAMES[species]
        self._id_field = STD_ID_DICT[species] if id is None else id

    @property
    def entity(self):
        """Name of the entity."""
        return "gene"

    @property
    def species(self):
        """The `common_name` of `Species` entity EntityTable."""
        return self._species

    @cached_property
    def df(self):
        """DataFrame."""
        if self.species not in {"human", "mouse"}:
            raise NotImplementedError
        else:
            if not self._filepath.exists():
                self._download_df()
            df = pd.read_feather(self._filepath)
            NormalizeColumns.gene(df, species=self.species)
            if not isinstance(df.index, pd.RangeIndex):
                df = df.reset_index().copy()
            return df.set_index(self._id_field)

    @cached_property
    def lookup(self):
        """Lookup object for auto-complete."""
        values = {i.replace("-", "_").rstrip("@"): i for i in self.df.index.values}
        nt = namedtuple(self._id_field, values.keys())

        return nt(**values)

    @check_datasetdir_exists
    def _download_df(self):
        from urllib.request import urlretrieve

        urlretrieve(
            f"https://bionty-assets.s3.amazonaws.com/{FILENAMES[self.species]}",
            self._filepath,
        )

    def curate(  # type: ignore
        self, df: pd.DataFrame, column: str = None
    ) -> pd.DataFrame:
        """Curate index of passed DataFrame to conform with default identifier.

        - If `column` is `None`, checks the existing index for compliance with
          the default identifier.
        - If `column` denotes an entity identifier, tries to map that identifier
          to the default identifier.

        Returns the DataFrame with the curated index and a boolean `__curated__`
        column that indicates compliance with the default identifier.
        """
        agg_col = ALIAS_DICT.get(self._id_field)
        df = df.copy()

        # if the query column name does not match any columns in the self.df
        # Bionty assume the query column and the self._id_field uses the same type of
        # identifier
        orig_column = column
        if column is not None and column not in self.df.columns:
            # normalize the identifier column
            column_norm = GENE_COLUMNS.get(column)
            if column_norm in df.columns:
                raise ValueError("{column_norm} column already exist!")
            else:
                column = self._id_field if column_norm is None else column_norm
                df.rename(columns={orig_column: column}, inplace=True)
            agg_col = ALIAS_DICT.get(column)
        return (
            super()
            .curate(df=df, column=column, agg_col=agg_col)
            .rename(columns={column: orig_column})
        )



from enum import Enum
from functools import cached_property
from pathlib import Path

import pandas as pd

from .._table import EntityTable

HERE = Path(__file__).parent
SPECIES_FILENAME = HERE / "tables/Species.csv"


class Field(str, Enum):
    """Species field names."""

    common_name = "common_name"
    scientific_name = "scientific_name"
    taxon_id = "taxon_id"


class Species(EntityTable):
    """Species.

    Args:
        id: Field name that should constitute the primary reference for each
            value. It will also be the primary key in the corresponding SQL EntityTable.
    """

    def __init__(self, id: Field = Field.common_name):
        self._id_field: Field = id  # type: ignore

    @cached_property
    def df(self) -> pd.DataFrame:
        """DataFrame."""
        # we want all columns to be read in as str
        # all numeric types in the EntityTable are versions & IDs
        # they behave like strings as, for instance, they cannot be added
        # if we wouldn't do this, we couldn't also properly aggregate in the groupby
        df = pd.read_csv(SPECIES_FILENAME, header=0, dtype=str)
        # we'll drop the display name as it's redundant with common_name
        df = df.drop("display_name", axis=1)
        # we'll lower case and _ concat the common name
        df.common_name = (
            df.common_name.str.lower()
            .str.translate({ord(c): "" for c in "!@#$%^&*()[]{};:,/<>?|`~=+'\""})
            .str.translate({ord(c): "_" for c in "-. "})
        )
        # we'll also drop nan as otherwise accession will raise a warning/error
        # there is a very small number of accession numbers that are nan
        df = df.dropna()
        # let's now do a groupby to get a unique index
        df = df.groupby(self._id_field).agg("; ".join)
        return df



from collections import namedtuple
from enum import Enum
from functools import cached_property

import pandas as pd

from ._logger import logger
from .dev._fix_index import (
    check_if_index_compliant,
    explode_aggregated_column_to_expand,
    get_compliant_index_from_column,
)


def _todict(x: list) -> dict:
    """Convert a list of strings to tab-completion allowed formats."""
    return {
        i.translate({ord(c): "_" for c in "-. !@#$%^&*()[]{};:,/<>?|`~=+'\""}): i
        for i in x
    }


class Field(str, Enum):
    field1 = "field1"
    field2 = "field2"


class EntityTable:
    """Biological entity as a EntityTable.

    See :doc:`tutorials` for background.
    """

    def __init__(self, id: Field = Field.field1):
        self._id_field = id

    @cached_property
    def df(self) -> pd.DataFrame:
        """DataFrame representation of EntityTable."""
        raise NotImplementedError

    @cached_property
    def lookup(self):
        """Return an auto-complete object for the bionty id."""
        values = _todict(self.df.index.to_list())
        nt = namedtuple("id", values.keys())

        return nt(**values)

    @cached_property
    def ontology(self):
        """Ontology."""
        return NotImplementedError

    def curate(
        self, df: pd.DataFrame, column: str = None, agg_col: str = None
    ) -> pd.DataFrame:
        """Curate index of passed DataFrame to conform with default identifier.

        - If `column` is `None`, checks the existing index for compliance with
          the default identifier.
        - If `column` denotes an entity identifier, tries to map that identifier
          to the default identifier.

        Returns the DataFrame with the curated index and a boolean `__curated__`
        column that indicates compliance with the default identifier.
        """
        df = df.copy()

        if agg_col is not None:
            # if provided a column with aggregated values, performs alias mapping
            alias_map = explode_aggregated_column_to_expand(
                self.df.reset_index(), aggregated_col=agg_col, target_col=self._id_field
            )[self._id_field]

        if column is None:
            # when column is None, use index as the input column
            index_name = df.index.name
            df["__mapped_index"] = (
                df.index if agg_col is None else df.index.map(alias_map)
            )
            df["orig_index"] = df.index
            df.index = df["__mapped_index"].fillna(df["orig_index"])
            del df["__mapped_index"]
            df.index.name = index_name
            matches = check_if_index_compliant(df.index, self.df.index)
        else:
            orig_series = df[column]
            df[column] = df[column] if agg_col is None else df[column].map(alias_map)
            df[column] = df[column].fillna(orig_series)
            new_index, matches = get_compliant_index_from_column(
                df=df,
                ref_df=self.df,
                column=column,
            )

            # keep the original index name as column name if exists
            # otherwise name it "orig_index"
            if df.index.name is None:
                df["orig_index"] = df.index
            else:
                df[df.index.name] = df.index
            df.index = new_index
            df.index.name = self._id_field
            df[column] = orig_series.values  # keep the original column untouched
        # annotated what complies with the default ID
        df["__curated__"] = matches
        # some stats for logging
        n_misses = len(matches) - matches.sum()
        frac_misses = round(n_misses / len(matches) * 100, 1)
        logger.warning(f"{n_misses} terms ({frac_misses}%) are not mappable.")
        return df



from functools import cached_property

import pandas as pd

from .._ontology import Ontology
from .._settings import settings
from .._table import EntityTable


class CellType(EntityTable):
    """Cell type.

    Edits of terms are coordinated and reviewed on:
    https://github.com/obophenotype/cell-ontology
    """

    def __init__(self, reload=False) -> None:
        self._reload = reload

    @cached_property
    def df(self) -> pd.DataFrame:
        """DataFrame."""
        return pd.DataFrame(
            [
                (term.id, term.name)
                for term in self.ontology.terms()
                if term.id.startswith("CL:")
            ],
            columns=["id", "name"],
        ).set_index("id")

    @cached_property
    def ontology(self):
        """Cell ontology."""
        url = "http://purl.obolibrary.org/obo/cl/cl-simple.obo"
        localpath = settings.dynamicdir / "cl-simple.obo"
        url = url if ((not localpath.exists()) or (self._reload)) else None
        ontology_ = Ontology(handle=localpath, url=url)
        if url is not None:
            ontology_.write_obo()
        return ontology_



from bionty import lookup


def test_lookup():
    assert lookup.gene_ids.ensembl_gene_id == "ensembl_gene_id"
    assert lookup.feature_model.gene == "gene"
    assert lookup.species.mouse == "mouse"
    assert lookup.cell_type.CL_0000057 == "CL:0000057"
    assert lookup.disease.MONDO_0000006 == "MONDO_0000006"



"""Bionty: mapping and standardizing biological entities.

Import the package::

   import bionty as bt

The base model for every entity class is:

.. autosummary::
   :toctree: .

   EntityTable

Entities:

.. autosummary::
   :toctree: .

   Species
   Gene
   Protein
   CellType
   Disease

Developer API:

.. autosummary::
   :toctree: .

   dev

"""

__version__ = "0.1.8"


# dynamic classes
from ._gene import Gene
from ._species import Species
from ._protein import Protein
from ._celltype import CellType
from ._disease import Disease

# tools
from ._normalize import NormalizeColumns
from ._table import EntityTable
from ._ontology import Ontology

# dev
from .dev import lookup



from collections import namedtuple
from typing import Iterable

from ... import CellType, Disease, Species
from ..._normalize import GENE_COLUMNS

FEATURES = ["gene", "protein"]


def lookup(values: Iterable[str]):
    """Look up a list of values via tab completion."""
    nt = namedtuple("feature", values)  # type: ignore
    return nt(**{i: i for i in values})


feature_model = lookup(values=FEATURES)
gene_ids = lookup(values=set(GENE_COLUMNS.values()))
species = Species().lookup
cell_type = CellType().lookup
disease = Disease().lookup



import pandas as pd
import pytest

from bionty.dev import get_compliant_index_from_column

data = {
    "hgnc_symbol": ["A1CF", "A1BG", "corrupted1", "corrupted2"],
    "hgnc_id": ["HGNC:24086", "HGNC:5", "corrupted1", "corrupted2"],
    "ensembl.gene_id": [
        "ENSG00000148584",
        "ENSG00000121410",
        "corrupted1",
        "corrupted2",
    ],
}
df = pd.DataFrame(data).set_index("hgnc_id")


def test_get_compliant_index_from_column():
    with pytest.raises(AssertionError):
        get_compliant_index_from_column(df, df, column="bla")
